@@ -1,7 +1,8 @@
 """
-This module provides functionality for parsing URLs for Dandisets & assets on
-Dandi Archive servers and for fetching the objects to which the URLs refer.
-See the docstring for `parse_dandi_url()` for a list of accepted URL formats.
+This module provides functionality for parsing URLs and other resource
+identifiers for Dandisets & assets on Dandi Archive servers and for fetching
+the objects to which the URLs refer.  See :ref:`resource_ids` for a list of
+accepted URL formats.
 
 Basic operation begins by calling `parse_dandi_url()` on a URL in order to
 acquire a `ParsedDandiURL` instance, which can then be used to obtain the
@@ -24,11 +25,13 @@ As a further convenience, a URL can be parsed and navigated in one fell swoop
 using the `navigate_url()` function.
 """
 
+from __future__ import annotations
+
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
 import re
 from time import sleep
-from typing import Iterator, Optional, Tuple
+from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, cast
 from urllib.parse import unquote as urlunquote
 
 from pydantic import AnyHttpUrl, BaseModel, parse_obj_as, validator
@@ -71,7 +74,7 @@ class ParsedDandiURL(ABC, BaseModel):
 
     @validator("api_url")
     def _validate_api_url(cls, v: AnyHttpUrl) -> AnyHttpUrl:
-        return parse_obj_as(AnyHttpUrl, v.rstrip("/"))
+        return cast(AnyHttpUrl, parse_obj_as(AnyHttpUrl, v.rstrip("/")))
 
     def get_client(self) -> DandiAPIClient:
         """
@@ -128,9 +131,9 @@ class ParsedDandiURL(ABC, BaseModel):
 
     @contextmanager
     def navigate(
-        self, strict: bool = False
+        self, *, strict: bool = False, authenticate: Optional[bool] = None
     ) -> Iterator[
-        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterator[BaseRemoteAsset]]
+        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
     ]:
         """
         A context manager that returns a triple of a
@@ -141,15 +144,30 @@ class ParsedDandiURL(ABC, BaseModel):
         If ``strict`` is true, then `get_dandiset()` is called with
         ``lazy=False`` and `get_assets()` is called with ``strict=True``; if
         ``strict`` is false, the opposite occurs.
+
+        If ``authenticate`` is true, then
+        `~dandi.dandiapi.DandiAPIClient.dandi_authenticate()` will be called on
+        the client before returning it.  If it is `None` (the default), the
+        method will only be called if the URL requires authentication (e.g., if
+        the resource(s) are embargoed).
+
+        .. versionchanged:: 0.35.0
+            ``authenticate`` added
         """
-        # We could later try to "dandi_authenticate" if run into permission
-        # issues.  May be it could be not just boolean but the "id" to be used?
         with self.get_client() as client:
-            yield (
-                client,
-                self.get_dandiset(client, lazy=not strict),
-                self.get_assets(client, strict=strict),
-            )
+            if authenticate:
+                client.dandi_authenticate()
+            try:
+                dandiset = self.get_dandiset(client, lazy=not strict)
+            except requests.HTTPError as e:
+                if e.response.status_code == 401 and authenticate is not False:
+                    lgr.info("Resource requires authentication; authenticating ...")
+                    client.dandi_authenticate()
+                    dandiset = self.get_dandiset(client, lazy=not strict)
+                else:
+                    raise
+            assets = self.get_assets(client, strict=strict)
+            yield (client, dandiset, assets)
 
 
 class DandisetURL(ParsedDandiURL):
@@ -162,9 +180,9 @@ class DandisetURL(ParsedDandiURL):
     ) -> Iterator[BaseRemoteAsset]:
         """Returns all assets in the Dandiset"""
         with _maybe_strict(strict):
-            yield from self.get_dandiset(client, lazy=not strict).get_assets(
-                order=order
-            )
+            d = self.get_dandiset(client, lazy=not strict)
+            assert d is not None
+            yield from d.get_assets(order=order)
 
 
 class SingleAssetURL(ParsedDandiURL):
@@ -202,6 +220,27 @@ class BaseAssetIDURL(SingleAssetURL):
         """Yields the ID of the asset (regardless of whether it exists)"""
         yield self.asset_id
 
+    @contextmanager
+    def navigate(
+        self, *, strict: bool = False, authenticate: Optional[bool] = None
+    ) -> Iterator[
+        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
+    ]:
+        with self.get_client() as client:
+            if authenticate:
+                client.dandi_authenticate()
+            dandiset = self.get_dandiset(client, lazy=not strict)
+            try:
+                assets = list(self.get_assets(client, strict=strict))
+            except requests.HTTPError as e:
+                if e.response.status_code == 401 and authenticate is not False:
+                    lgr.info("Resource requires authentication; authenticating ...")
+                    client.dandi_authenticate()
+                    assets = list(self.get_assets(client, strict=strict))
+                else:
+                    raise
+            yield (client, dandiset, assets)
+
 
 class AssetIDURL(SingleAssetURL):
     """
@@ -219,7 +258,9 @@ class AssetIDURL(SingleAssetURL):
         nothing is yielded if ``strict`` is false.
         """
         with _maybe_strict(strict):
-            yield self.get_dandiset(client, lazy=not strict).get_asset(self.asset_id)
+            d = self.get_dandiset(client, lazy=not strict)
+            assert d is not None
+            yield d.get_asset(self.asset_id)
 
     def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
         """Yields the ID of the asset (regardless of whether it exists)"""
@@ -236,9 +277,9 @@ class AssetPathPrefixURL(MultiAssetURL):
     ) -> Iterator[BaseRemoteAsset]:
         """Returns the assets whose paths start with `path`"""
         with _maybe_strict(strict):
-            yield from self.get_dandiset(
-                client, lazy=not strict
-            ).get_assets_with_path_prefix(self.path, order=order)
+            d = self.get_dandiset(client, lazy=not strict)
+            assert d is not None
+            yield from d.get_assets_with_path_prefix(self.path, order=order)
 
 
 class AssetItemURL(SingleAssetURL):
@@ -260,6 +301,7 @@ class AssetItemURL(SingleAssetURL):
         """
         try:
             dandiset = self.get_dandiset(client, lazy=not strict)
+            assert dandiset is not None
             # Force evaluation of the version here instead of when
             # get_asset_by_path() is called so we don't get nonexistent
             # dandisets with unspecified versions mixed up with nonexistent
@@ -306,9 +348,9 @@ class AssetFolderURL(MultiAssetURL):
         if not path.endswith("/"):
             path += "/"
         with _maybe_strict(strict):
-            yield from self.get_dandiset(
-                client, lazy=not strict
-            ).get_assets_with_path_prefix(path, order=order)
+            d = self.get_dandiset(client, lazy=not strict)
+            assert d is not None
+            yield from d.get_assets_with_path_prefix(path, order=order)
 
 
 @contextmanager
@@ -322,9 +364,9 @@ def _maybe_strict(strict: bool) -> Iterator[None]:
 
 @contextmanager
 def navigate_url(
-    url: str, strict: bool = False
+    url: str, *, strict: bool = False, authenticate: Optional[bool] = None
 ) -> Iterator[
-    Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterator[BaseRemoteAsset]]
+    Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
 ]:
     """
     A context manager that takes a URL pointing to a DANDI Archive and
@@ -333,17 +375,29 @@ def navigate_url(
     identified in the URL (if any), and the assets specified by the URL (or, if
     no specific assets were specified, all assets in the Dandiset).
 
+    .. versionchanged:: 0.35.0
+        ``authenticate`` added
+
     :param str url: URL which might point to a Dandiset, folder, or asset(s)
     :param bool strict:
         If true, then `get_dandiset()` is called with ``lazy=False`` and
         `get_assets()` is called with ``strict=True``; if false, the opposite
         occurs.
+    :param Optional[bool] authenticate:
+        If true, then `~dandi.dandiapi.DandiAPIClient.dandi_authenticate()`
+        will be called on the client before returning it.  If `None` (the
+        default), the method will only be called if the URL requires
+        authentication (e.g., if the resource(s) are embargoed).
     :returns: Context manager that yields a ``(client, dandiset, assets)``
         tuple; ``client`` will have a session established for the duration of
         the context
     """
     parsed_url = parse_dandi_url(url)
-    with parsed_url.navigate(strict=strict) as (client, dandiset, assets):
+    with parsed_url.navigate(strict=strict, authenticate=authenticate) as (
+        client,
+        dandiset,
+        assets,
+    ):
         yield (client, dandiset, assets)
 
 
@@ -354,7 +408,7 @@ class _dandi_url_parser:
     dandiset_id_grp = f"(?P<dandiset_id>{DANDISET_ID_REGEX})"
     # Should absorb port and "api/":
     server_grp = "(?P<server>(?P<protocol>https?)://(?P<hostname>[^/]+)/(api/)?)"
-    known_urls = [
+    known_urls: List[Tuple[re.Pattern[str], Dict[str, Any], str]] = [
         # List of (regex, settings, display string) triples
         #
         # Settings:
@@ -479,57 +533,8 @@ class _dandi_url_parser:
     @classmethod
     def parse(cls, url: str, *, map_instance: bool = True) -> ParsedDandiURL:
         """
-        Parse a Dandi Archive URL and return a `ParsedDandiURL` instance.
-
-        The accepted URL formats are as follows.  Text in [brackets] is
-        optional.  A ``server`` field is a base API, GUI, or redirector URL for
-        a registered Dandi Archive instance.  If an optional ``version`` field
-        is omitted from a URL, the given Dandiset's most recent published
-        version, if any, or else its draft version will be used.
-
-        - :samp:`https://identifiers.org/DANDI:{dandiset-id}[/{version}]`
-          (case insensitive; ``version`` cannot be "draft") when it redirects
-          to one of the other URL formats
-
-        - :samp:`DANDI:{dandiset-id}[/{version}]` — (case insensitive;
-          ``version`` cannot be "draft") Abbreviation for the above
-
-        - Any ``https://dandiarchive.org/`` or
-          ``https://*dandiarchive-org.netflify.app/`` URL which redirects to
-          one of the other URL formats
-
-        - :samp:`https://{server}[/api]/[#/]dandiset/{dandiset-id}[/{version}][/files[?location={path}]]`
-          — If ``path`` is not specified, the URL refers to a Dandiset and is
-          converted to a `DandisetURL`.  If ``path`` is specified and it ends
-          with a forward slash, the URL refers to an asset folder and is
-          converted to an `AssetFolderURL` instance; if it does not end with a
-          slash, it refers to a single asset and is converted to an
-          `AssetItemURL` instance.
-
-        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}[/versions[/{version}]]`
-          — Refers to a Dandiset and is converted to a `DandisetURL`
-
-        - :samp:`https://{server}[/api]/assets/{asset-id}[/download]` — Refers
-          to a single asset by identifier and is converted to a
-          `BaseAssetIDURL`
-
-        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}/versions/{version}/assets/{asset-id}[/download]`
-          — Refers to a single asset and is converted to an `AssetIDURL`
-
-        - :samp:`https://{server}[/api]/dandisets/{dandiset-id}/versions/{version}/assets/?path={path}`
-          — Refers to all assets in the given Dandiset whose paths begin with
-          the prefix ``path``; converted to an `AssetPathPrefixURL`
-
-        - :samp:`dandi://{instance-name}/{dandiset-id}[@{version}][/{path}]`,
-          where ``instance-name`` is the name of a registered Dandi Archive
-          instance.  If ``path`` is not specified, the URL refers to a Dandiset
-          and is converted to a `DandisetURL`.  If ``path`` is specified and it
-          ends with a forward slash, the URL refers to an asset folder and is
-          converted to an `AssetFolderURL` instance; if it does not end with a
-          slash, it refers to a single asset and is converted to an
-          `AssetItemURL` instance.
-
-        - Any other HTTPS URL that redirects to one of the above
+        Parse a Dandi Archive URL and return a `ParsedDandiURL` instance.  See
+        :ref:`resource_ids` for the supported URL formats.
 
         :raises UnknownURLError: if the URL is not one of the above
         """
@@ -576,7 +581,10 @@ class _dandi_url_parser:
                             )
                         )
                     known_instance = get_instance(settings["map_instance"])
-                    parsed_url.api_url = known_instance.api
+                    assert known_instance.api is not None
+                    parsed_url.api_url = cast(
+                        AnyHttpUrl, parse_obj_as(AnyHttpUrl, known_instance.api)
+                    )
                 continue  # in this run we ignore and match further
             elif "instance_name" in groups:
                 try:
